;ATTiny2313 with external 4.194304 MHz crystal
;Fuses: low = 0xDD, high = 0xDB, ext = 0xFF

;simple quartz clock for 6 7-segment LED displays (multiplexed)
;(C) Arne Rossius 2008-08-06, http://www.ebps.de.vu/
;[based on Tiny24 simple quartz clock]

;PORTS:
;  PB0   = Anode for 1's of seconds
;  PB1   = Anode for 10's of seconds
;  PB2   = Anode for 1's of minutes
;  PB3   = Anode for 10's of minutes
;  PB4   = Anode for 1's of hours
;  PB5   = Anode for 10's of hours
;  PB6   = "Minutes" key (to ground)
;  PB7   = "Hours" key (to ground)
;  PD0-6 = Segments a to g

.equ	KEY_REPEATSTART = 16 ;16/16s = 1s until repeating starts
.equ	KEY_REPEAT = 5 ;16/5 = 3.2 repeats per second

.include "../include/tn2313def.inc"
 
.def	disp1 = R0
.def	disp2 = R1
.def	disp3 = R2
.def	disp4 = R3
.def	disp5 = R4
.def	disp6 = R5
 
.def	temp = R16
.def	temp2 = R17
.def	sec1 = R18
.def	sec10 = R19
.def	min1 = R20
.def	min10 = R21
.def	hr1 = R22
.def	hr10 = R23
.def	disp = R24
.def	repeat = R25

;R31:R30: Z pointer for Flash access

;===============================================================================

.cseg
.org 0x000
	rjmp	reset
.org OVF0addr
	rjmp	keyrepeat
.org PCINTaddr
	rjmp	keytoggle

;===============================================================================

disptab:
	.db 0x3f, 0x06, 0x5b, 0x4f, 0x66, 0x6d, 0x7d, 0x07
	.db 0x7f, 0x6f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00

;===============================================================================

keytoggle:
	;pin change interrupt for hour/minute keys
	push	temp
	in	temp, SREG
	push	temp
	
	in	temp, PINB
	andi	temp, 0xC0
	cpi	temp, 0xC0
	breq	key_released
	
	;key pressed: reset and enable debounce/repeat timer
	ldi	temp, 128
	out	TCNT0, temp
	ldi	temp, 0x04 ;Clk/256 => ~8ms debouncing time
	out	TCCR0B, temp
	
keytoggle_end:
	;return
	pop	temp
	out	SREG, temp
	pop	temp
	reti
	
key_released:
	;key released: disable debounce/repeat timer
	ldi	temp, 0x00
	out	TCCR0B, temp
	ldi	repeat, 0
	rjmp	keytoggle_end

;===============================================================================

keyrepeat:
	;key debounce/repeat timer overflow
	push	temp
	in	temp, SREG
	push	temp
	
	;change from debounce to repeat mode
	cpi	repeat, 0
	brne	repeatdiv
	ldi	temp, 0x05 ;Clk/1024
	out	TCCR0B, temp
	ldi	temp, 0
	out	TCNT0, temp
	ldi	repeat, KEY_REPEATSTART
	rjmp	key_action
	
repeatdiv:
	;repeated keypress
	dec	repeat
	brne	keyrepeat_end
	ldi	repeat, KEY_REPEAT
key_action:
	;perform key action
	sbis	PINB, 7
	rjmp	key_hr
	sbis	PINB, 6
	rjmp	key_min
	
keyrepeat_end:
	;return
	pop	temp
	out	SREG, temp
	pop	temp
	reti
	
key_hr:
	;"Hour" key: increment hours
	rcall	inc_hr
	rcall	update_disp
	rjmp	keyrepeat_end

key_min:
	;"Minute" key: increment minutes and reset seconds
	rcall	inc_min
	clr	sec10
	clr	sec1
	rcall	update_disp
	ldi	temp, 0
	out	TCNT1H, temp
	out	TCNT1L, temp
	rjmp	keyrepeat_end
	
;===============================================================================

reset:
	;stack pointer setup
	ldi	temp, RAMEND
	out	SPL, temp
	
	;port setup
	ldi	temp, 0x3F
	out	DDRB, temp
	ldi	temp, 0xFF
	out	PORTB, temp
	ldi	temp, 0x7F
	out	DDRD, temp
	ldi	temp, 0x7F
	out	PORTD, temp
	
	;Timer 0 setup (Key debounce/repeat timer)
	ldi	temp, 1<<TOIE0
	out	TIMSK, temp
	
	;Timer 1 setup (Clock timer)
	ldi	temp, 1<<WGM12 | 0x05 ;CTC Mode, Clk/1024
	out	TCCR1B, temp
	ldi	temp, HIGH(4096) ;overflow every second (4194304 Hz / (1024 * 4096) = 1 Hz)
	out	OCR1AH, temp
	ldi	temp, LOW(4096)
	out	OCR1AL, temp
	
	;Pin Change interrupt setup
	ldi	temp, 0xC0
	out	PCMSK, temp
	ldi	temp, 1<<PCIE
	out	GIMSK, temp
	
	;register setup
	ldi	hr10, 0
	ldi	hr1, 0
	ldi	min10, 0
	ldi	min1, 0
	ldi	sec10, 0
	ldi	sec1, 0
	rcall	update_disp
	
	;enable interrupts
	sei
	
;-------------------------------------------------------------------------------
	
loop:
	ldi	disp, 0x20
	
muxloop:
	;increment clock if timer reached 1s
	in	temp, TIFR
	sbrs	temp, OCF1A
	rjmp	timeinc_end
	ldi	temp, 1<<OCF1A ;reset flag
	out	TIFR, temp
	rcall	inc_sec
	rcall	update_disp
timeinc_end:

	;get next display data
	sbrc	disp, 5
	mov	temp, disp1
	sbrc	disp, 4
	mov	temp, disp2
	sbrc	disp, 3
	mov	temp, disp3
	sbrc	disp, 2
	mov	temp, disp4
	sbrc	disp, 1
	mov	temp, disp5
	sbrc	disp, 0
	mov	temp, disp6
	com	temp

	;calculate next display anode selection mask
	com	disp
	
	;switch to next display
	ldi	temp2, 0xFF
	out	PORTB, temp2
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	out	PORTD, temp
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	out	PORTB, disp
	
	;restore value of 'disp'
	com	disp
	
	;delay
	ldi	temp, 5
	ldi	temp2, 0
mux_delay:
	dec	temp2
	brne	mux_delay
	dec	temp
	brne	mux_delay
	
	lsr	disp
	brne	muxloop
	rjmp	loop
	
;===============================================================================

inc_sec:
	;increment seconds
	inc	sec1
	cpi	sec1, 10
	brlo	inc_end
	clr	sec1
	inc	sec10
	cpi	sec10, 6
	brlo	inc_end
	clr	sec10
inc_min:
	;increment minutes
	inc	min1
	cpi	min1, 10
	brlo	inc_end
	clr	min1
	inc	min10
	cpi	min10, 6
	brlo	inc_end
	clr	min10
inc_hr:
	;increment hours
	inc	hr1
	cpi	hr1, 10
	brlo	hr_24
	clr	hr1
	inc	hr10
	rjmp	inc_end
hr_24:
	cpi	hr1, 4
	brne	inc_end
	cpi	hr10, 2
	brne	inc_end
	clr	hr1
	clr	hr10
inc_end:
	ret
	
;===============================================================================

update_disp:
	;convert time to 7-segment display representation
	ldi	ZH, 0
	ldi	ZL, LOW(disptab*2)
	add	ZL, hr10
	lpm	disp1, Z
	ldi	ZL, LOW(disptab*2)
	add	ZL, hr1
	lpm	disp2, Z
	ldi	ZL, LOW(disptab*2)
	add	ZL, min10
	lpm	disp3, Z
	ldi	ZL, LOW(disptab*2)
	add	ZL, min1
	lpm	disp4, Z
	ldi	ZL, LOW(disptab*2)
	add	ZL, sec10
	lpm	disp5, Z
	ldi	ZL, LOW(disptab*2)
	add	ZL, sec1
	lpm	disp6, Z
	ret